# coding=utf-8

import sys
import logging
from sinagot.utils import (
    record_log_file_path,
    LOG_ORIGIN,
    LOG_RECORD_ID,
    LOG_STEP_LABEL,
)

LOG_FORMAT = "%(asctime)s | %(task)s-%(modality)s-%({})s : %(message)s".format(
    LOG_STEP_LABEL
)


def logger_factory(config):
    """Create logger with config info"""

    config = config["log"]

    logger = logging.getLogger(config["name"])

    handlers = list(logger.handlers)
    for hdlr in handlers:
        logger.removeHandler(hdlr)

    handler = logging.StreamHandler(sys.stdout)
    if "format" in config:
        formatter = logging.Formatter(config["format"])
        handler.setFormatter(formatter)
    handler.addFilter(filter_no_script)
    logger.addHandler(handler)
    if "level" in config:
        logger.setLevel(config["level"])
    logger.debug("log initialized")

    script_handler = logging.StreamHandler(sys.stdout)
    script_handler.setFormatter(
        logging.Formatter("%({})s | ".format(LOG_RECORD_ID) + LOG_FORMAT)
    )
    script_handler.addFilter(filter_is_script)
    logger.addHandler(script_handler)

    return logger


def filter_no_script(record):
    try:
        return record.__dict__[LOG_ORIGIN] != "script"
    except KeyError:
        return True


def filter_is_script(record):
    try:
        return record.__dict__[LOG_ORIGIN] == "script"
    except KeyError:
        return False



# coding=utf-8

from pathlib import Path
from importlib import import_module, util as importutil


class Model:
    """Base class for every model except scripts"""

    _REPR_ATTRIBUTES = []
    """List of instance attributes to display in __repr__"""
    dataset = None
    """Master dataset"""
    config = None
    """Config dic of the dataset"""
    logger = None
    """Logger object"""

    def __init__(self, dataset):
        self.dataset = dataset
        self.config = dataset.config
        self.logger = dataset.logger

    def __repr__(self):
        attributes = ", ".join(
            [
                "{label}: {value}".format(label=label, value=getattr(self, label))
                for label in self._REPR_ATTRIBUTES
                if hasattr(self, label)
            ]
        )
        return "<{} instance | ".format(self.__class__.__name__) + attributes + ">"

    def _get_module(self, class_name, *args):
        """Import a module from scripts folder"""

        path = Path(self.dataset._scripts_path, *args)
        spec = importutil.spec_from_file_location(
            ".".join(args[-3:]), path.with_suffix(".py")
        )
        module = importutil.module_from_spec(spec)
        spec.loader.exec_module(module)
        return getattr(module, class_name)



# coding=utf-8

from typing import Optional, Generator, TypeVar
import pandas as pd
from sinagot.models import Model, StepCollection


class Scope(Model):
    """
    A Scope instance can :
    
    - Handle task and modality subscopes.

    - Handle StepCollection with `.steps` attribute.

    Note:
        Scope is base class for [Subset](subset.md) and [Record](record.md).
    """

    _REPR_ATTRIBUTES = ["task", "modality"]
    _MODEL_TYPE = None
    _subscope_class = None
    _tasks = []
    task = None
    _modalities = []
    modality = None

    def __init__(
        self,
        dataset: "Dataset",
        task: Optional[str] = None,
        modality: Optional[str] = None,
    ):
        """
        Args:
            dataset: Root Dataset.
            task: Task of the scope.
            modality: Modality of the scope.
        """

        dataset.logger.debug(
            "## init scope for %s. _subscope_class: %s", self, self._subscope_class
        )
        if task is not None:
            self.task: str = task
            """Task of the scope. If `None`, the scope represents all available tasks."""
        if modality is not None:
            self.modality: str = modality
            """Modality of the scope. If `None`, the scope represents all available modalities."""
        if self._subscope_class is None:
            self._subscope_class = self.__class__
        super().__init__(dataset)

        self.steps: "StepCollection" = self._set_step_collection()
        """Collection of steps."""

    @classmethod
    def _set_subscopes(cls, dataset):
        """Create subscope tasks and modalities attributes
        Run once at dataset initialization"""

        SUBSCOPES = (
            ("task", "tasks"),
            ("modality", "modalities"),
        )

        for subscope, collection in SUBSCOPES:
            setattr(
                cls,
                "_{}".format(collection),
                [
                    cls._add_subscope(subscope=subscope, value=value, dataset=dataset)
                    for value in dataset.config[collection].keys()
                ],
            )

    @classmethod
    def _add_subscope(cls, subscope, value, dataset):
        """Add subscope to self.__class__ as a property
        that call an instance of local defined subclass"""

        # Search for custom subscope class in config
        model_type = cls._MODEL_TYPE
        _subscope_class = cls
        if subscope == "modality":
            try:
                class_name = dataset.config["modalities"][value]["models"][model_type]
                _subscope_class = dataset._get_module(
                    class_name, value, "models", model_type
                )
            except KeyError:
                pass

        setattr(
            cls,
            value,
            property(cls._property_factory(_subscope_class, subscope, value)),
        )
        return value

    @staticmethod
    def _property_factory(_subscope_class, subscope, value):
        def get(self):
            if self._is_valid_subscope(subscope, value):
                kwargs = {
                    "dataset": self.dataset,
                    "task": self.task,
                    "modality": self.modality,
                }
                kwargs[subscope] = value
                if self._MODEL_TYPE == "record":
                    kwargs["record_id"] = self.id
                return _subscope_class(**kwargs)
            else:
                raise AttributeError(
                    "{} {} is not valid for {}".format(subscope, value, self)
                )

        return get

    def _is_valid_subscope(self, subscope, value):
        if getattr(self, subscope) is not None:
            return False
        if subscope == "task":
            return self.modality in self.dataset.config["tasks"][value][
                "modalities"
            ] + [None]
        elif subscope == "modality":
            return (
                self.task is None
                or value in self.dataset.config["tasks"][self.task]["modalities"]
            )
        else:
            return False

    @property
    def is_unit(self) -> bool:
        """Check if has both task and modality not `None`."""

        return not (self.task is None or self.modality is None)

    def units(self) -> Generator["Scope", None, None]:
        """
        Generate each 'unit' subscopes of the current scope,
        i.e. subscope with specific task and modality

        Returns:
            Subscope with a task attribute.
        """

        if self.task is None:
            tasks = self.tasks()
        else:
            tasks = [self]
        for task in tasks:
            if task.modality is None:
                modalities = task.modalities()
            else:
                modalities = [task]
            for modality in modalities:
                yield modality

    def tasks(self) -> Generator["Scope", None, None]:
        """
        Generate subscopes of the current scope for each of its task.

        Returns:
            Subscope with a task attribute.
        """

        for task in self._tasks:
            if self._is_valid_subscope("task", task):
                model = getattr(self, task)
                yield model

    def modalities(self) -> Generator["Scope", None, None]:
        """
        Generate subscopes of the current scope for each of its modalities.

        Returns:
            Subscope with a modality attribute.
        """

        for modality in self._modalities:
            if self._is_valid_subscope("modality", modality):
                yield getattr(self, modality)

    def _set_step_collection(self):
        """Used in __init__(), set 'steps' attribute to an instance of step collection class"""

        if self.is_unit:
            step_collection_class = StepCollection
            return step_collection_class(self)

    def run(self, step_label: Optional[str] = None, force: Optional[bool] = False):
        """Run all steps of the scope.
        
        Args:
            step_label: if not `None`, run only for the step with this label.
            force: Force run and overwrites result file(s) if already exist(s).
        """

        return self.dataset._run_manager.run(self)

    def status(self) -> pd.DataFrame:
        """
        Returns:
            Status for each step.
        """

        try:
            return pd.concat(
                [
                    unit.steps.status()
                    for unit in self.units()
                    if unit.steps.status() is not None
                ]
            )
        except ValueError:
            return pd.DataFrame()



# coding=utf-8

import inspect
import pandas as pd
from sinagot.models import Model
from sinagot.utils import StepStatus, LOG_STEP_LABEL, LOG_STEP_STATUS


class Step(Model):

    _REPR_ATTRIBUTES = ["scope", "task", "modality", "label"]
    _MODEL_TYPE = "step"
    label = None
    id = None

    def __init__(self, script, model):

        self.model = model
        self.task = model.task
        self.modality = model.modality
        super().__init__(model.dataset)

        if inspect.isclass(script):
            script_class = script
            self.label = script.__class__.__name__
        elif isinstance(script, str):
            script_class = self._get_module("Script", self.modality, script)
            self.label = script
        else:
            raise AttributeError("Type {} is not valid for script".format(type(script)))

        if model._MODEL_TYPE == "record":
            self.id = model.id

        self.script_class = script_class
        self.script = script_class(
            data_path=self.dataset._data_path,
            id_=self.id,
            task=self.task,
            logger_namespace=self.logger.name,
        )

    def status(self):
        if self.script.path.output.exists():
            return StepStatus.DONE
        try:
            logs = self.logs()
            status = logs[logs[LOG_STEP_STATUS].notna()].iloc[0][LOG_STEP_STATUS]
        except:
            status = None
        if status == StepStatus.PROCESSING:
            return StepStatus.PROCESSING
        if self.script.path.input.exists():
            return StepStatus.DATA_READY
        else:
            return StepStatus.INIT

    def run(self):
        self.model.run()

    def logs(self):
        try:
            return self.model.logs().query(
                "{} == '{}'".format(LOG_STEP_LABEL, self.label)
            )
        except:
            return pd.DataFrame()



from sinagot.models import Model, Step, Record


class RunManager(Model):
    """Manage multiple run in sequential or parallel mode"""

    kwargs = {}

    def run(self, model, **kwargs):

        self.kwargs = kwargs

        if model._MODEL_TYPE == "record":
            records = [model]
        else:
            records = model.all()

        return self._run(records)

    @staticmethod
    def record_structure(record):
        return record._structure_status(("record_id", "task", "modality", "step_label"))

    def _run(self, records):

        for record in records:
            shape = self.record_structure(record)
            for task in shape.values():
                for modality in task.values():
                    for step in modality:
                        self.run_step_factory(step)(None)

    def run_step_factory(self, step):
        id_ = step["record_id"]
        task = step["task"]
        modality = step["modality"]
        step_label = step["step_label"]
        script_class = self._get_module("Script", modality, step_label)
        script = script_class(
            data_path=self.dataset._data_path,
            id_=id_,
            task=task,
            logger_namespace=self.logger.name,
        )

        def func(arg):
            # if to_run:
            script._run()
            return arg

        return func



# coding=utf-8

import json
from collections import defaultdict
from typing import Optional
from io import StringIO
import pandas as pd
from sinagot.models import Scope
from sinagot.utils import record_log_file_path
from sinagot.utils import (
    LOG_STEP_LABEL,
    LOG_STEP_STATUS,
)


class Record(Scope):
    """
    A Record instance is used to manipulate a single record data.
    It's accessed from a [Subset](subset.md).
    
    Note:
        Inherite [Scope](scope.md) methods.

    Example:

    ```python
    rec = ds.get("RECORD-ID") # ds is a Dataset instance
    ```
    """

    _REPR_ATTRIBUTES = ["id", "task", "modality"]
    _MODEL_TYPE = "record"

    def __init__(self, dataset: "Dataset", record_id: str, *args, **kwargs):
        """
        Args:
            dataset: Root Dataset.
            record_id: ID of the record.
        """

        # TODO: raise exception if id not match regex in config
        self.id = record_id
        super().__init__(dataset, *args, **kwargs)

    @property
    def subset(self) -> "Subset":
        """
        Subset instance with same scope than the record i.e. the same task and modality value.
        """

        subset = self.dataset
        if self.task is not None:
            subset = getattr(subset, self.task)
        if self.modality is not None:
            subset = getattr(subset, self.modality)
        return subset

    def exists(self) -> bool:
        """
        Check if the record exists in the subset with the same task and modality values.

        Returns:
            True if the record exists
        """

        return self.subset.has(self.id)

    def logs(self) -> pd.DataFrame:
        """
        Returns:
            Logs history.
        """

        log_path = record_log_file_path(self.dataset._data_path, self.id)
        if log_path.exists():
            json = "[{}]".format(",".join(log_path.read_text().split("\n")[:-1]))
            df = pd.read_json(StringIO(json), orient="records")
            for key in ("task", "modality"):
                k = getattr(self, key)
                if k is not None:
                    df = df.query("{} == '{}'".format(key, k))
            return df.sort_values("time", ascending=False)
        else:
            return pd.DataFrame()

    # TODO: Test

    def _count_detail(self):

        if self.is_unit:
            return pd.Series(
                {
                    "task": self.task,
                    "modality": self.modality,
                    "count": sum([self.exists()]),
                }
            )
        else:
            return pd.DataFrame([unit._count_detail() for unit in self.units()])

    def count_detail(
        self, groupby: Optional[str] = None, group_mode: Optional[str] = "all"
    ) -> pd.DataFrame:
        """
        Args:
            groupby: If not `None`, columns to perform a groupby.
            group_mode: method to count the groupby. Possible values:

                - 'all': Return 1 if exists in all units of the aggregation, 0 if not.

                - 'any': Return 1 if exists in at least one unit.

                - 'mean': Return an mean for each unit.

        Returns:
            Detail count if record exists in each of tasks and modality.

        Raises:
            AttributeError: if group_mode not in ('all', 'any', 'mean').
        """

        if isinstance(groupby, str):
            groupby = [groupby]

        if group_mode not in ("all", "any", "mean"):
            raise AttributeError("aggregate_mode arg not valid")

        count = self._count_detail()
        if self.is_unit:
            count = pd.DataFrame([count])
        base = self._count_backbone()
        count = (
            base.merge(count, on=["task", "modality"])
            .eval("count = count_x + count_y")
            .reindex(["task", "modality", "count"], axis=1)
        )
        if groupby is not None:
            count = count.groupby(groupby)
            if group_mode == "all":
                count = count.min()
            if group_mode == "any":
                count = count.max()
            if group_mode == "mean":
                count = count.mean()
            return count.reset_index().reindex(groupby + ["count"], axis=1)
        return count

    def _count_backbone(self):
        """Used for count_detail() method."""

        return pd.DataFrame(
            [
                pd.Series({"task": task, "modality": modality, "count": 0})
                for task in self.config["tasks"]
                for modality in self.config["modalities"]
            ]
        )

    def _structure_status(self, columns):

        df = self.status().sort_values(["step_index"])

        def status_default_dict():
            def dict_to_list():
                return defaultdict(list)

            return defaultdict(dict_to_list)

        def get_status_dict(df):
            sd = status_default_dict()
            for row in df.itertuples():
                sd[row.task][row.modality].append(
                    {key: getattr(row, key) for key in columns}
                )
            return sd

        return get_status_dict(df)

    def status_json(self) -> str:
        """
        Returns:
            Status in JSON format for web API.
        """

        columns = ("step_index", LOG_STEP_LABEL, LOG_STEP_STATUS)

        return json.dumps(self._structure_status(columns))



import json
import asyncio
from dask.distributed import LocalCluster, Client, Future
from sinagot.models import Record, RunManager
from sinagot.config import ConfigurationError


class DaskRunManager(RunManager):
    """Manage multiple run in sequential or parallel mode"""

    def __init__(self, dataset):
        super().__init__(dataset)
        self.dask_config = dataset.config.get("dask", {})
        scheduler_config = self.dask_config.get("scheduler", {})
        mode = scheduler_config.pop("mode", "local")
        if mode == "local":
            cluster = LocalCluster(**scheduler_config)
            dataset.scheduler_address = cluster.scheduler_address
        elif mode == "distributed":
            dataset.scheduler_address = scheduler_config["scheduler_address"]
        else:
            raise ConfigurationError("{} model is not enable for dask".format(mode))
        self.client = Client(cluster.scheduler_address)

    def _run(self, records):

        futures = [
            self.dask_futures(self.record_structure(record)) for record in records
        ]

        asynchronous = self.dask_config.get("asynchronous", False)
        return self.client.gather(futures, asynchronous=asynchronous)

    def dask_futures(self, target):
        if isinstance(target, dict):
            return self.dask_parallel_futures(target)
        else:
            return self.dask_steps_future(target)

    def dask_parallel_futures(self, collection):
        def func(*args):
            return args

        return self.client.map(
            func, [self.dask_futures(item) for item in collection.values()]
        )

    def dask_steps_future(self, collection):

        future = None

        for item in collection:
            func = self.run_step_factory(item)
            future = self.client.submit(func, future)

        return future



# coding=utf-8

import os
import re
from typing import Generator
import pandas as pd
from sinagot.models import Scope, Record


class Subset(Scope):
    """
    A Subset is used to manipulate a collection of [Record](record.md).

    Note:
        Inherite [Scope](scope.md) methods.

    Example:

    ```python
    # access subset of all EEG records (EEG is a task)
    sub = ds.EEG  # ds is a Dataset instance
    ```
    """

    _MODEL_TYPE = "subset"
    _record_class = Record

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        modality = self.modality
        # Check for custom record class
        if modality is not None:
            try:
                class_name = self.config["modalities"][modality]["models"]["record"]
                self._record_class = self._get_module(
                    class_name, modality, "models", "record"
                )
            except KeyError:
                pass

    def ids(self) -> Generator[str, None, None]:
        """Generator all record ids within the subset.
        
        Returns:
            Record ID.
        """
        # TODO: use dataset cache for ids
        if self.is_unit:
            for record_id in self._ids_unit():
                yield record_id
        else:
            units = self.units()
            ids = []
            for unit in units:
                for record_id in unit.ids():
                    if not record_id in ids:
                        ids.append(record_id)
                        yield record_id

    def _ids_unit(self):
        """Return the list of all records ids i.e. subfolder names"""
        ids = []
        config = self.config
        try:
            file_match = config["modalities"][self.modality]["file_match"]
        except KeyError:
            file_match = config["records"]["file_match"]

        try:
            first_script = self.steps.first().script
        except IndexError:
            first_script = False
        if first_script:
            pattern = os.path.join(*first_script.PATH_IN)
            path_ = first_script.PATH_IN
            if not file_match[0]:
                path_ = path_[:-1]
            pattern = os.path.join(*path_)
            pattern = pattern.format(
                id="({})".format(config["records"]["id_pattern"]),
                task="(?:{})".format(self.task),
            )
            root_path = os.path.join(self.dataset._data_path, first_script.PATH_IN[0])
            for root, dirs, files in os.walk(root_path):
                if file_match[1]:
                    # Search ID wihtin files
                    item_set = files
                else:
                    # Search ID wihtin dirs
                    item_set = dirs
                for item in item_set:
                    path = os.path.join(root, item)
                    m = re.search(pattern, path)
                    if m and (len(m.groups()) > 0):
                        record_id = m.group(1)
                        if record_id not in ids:
                            ids.append(record_id)
                            yield record_id

    def all(self) -> Generator["Record", None, None]:
        """Generate all records instances of the subset.
        
        Returns:
            sinagot.models.Record: Record instance
        """

        for record_id in self.ids():
            yield self.get(record_id)

    def first(self) -> "Record":
        """
        Get the first record found.
        
        Returns:
            Record instance
        """

        for record_id in self.ids():
            return self.get(record_id)

    def get(self, record_id: str) -> "Record":
        """
        Get record by ID.

        Args:
            record_id: record ID

        Returns:
            Record instance
        """

        return self._record_class(
            dataset=self.dataset,
            record_id=record_id,
            task=self.task,
            modality=self.modality,
        )

    def has(self, record_id: str) -> bool:
        """
        Check existance of a record in the subset.
        
        Args:
            record_id: Id of the record.
        
        Returns:
            True if record exists.
        """

        for id_ in self.ids():
            if id_ == record_id:
                return True
        return False

    def count(self) -> int:
        """
        Count all records.

        Returns:
            Number of records.
        """

        return sum(1 for rec in self.ids())

    # TODO: To test

    def count_detail(self, *args, **kwargs) -> pd.DataFrame:
        """
        Returns:
            Number of records present in each tasks and modalities.

        Note:
            Refer to [`Record.count_detail()`](record.md#sinagot.models.record.Record.count_detail) for more information
        """

        count = None

        self.logger.debug("Begin count_detail")
        for rec in self.all():
            self.logger.debug("Begin count_detail for record {}".format(rec.id))
            try:
                count_ = rec.count_detail(*args, **kwargs)
                if count is None:
                    count = count_
                    columns = count.columns
                else:
                    count = count.append(count_)
                    if "count_x" in count.columns:
                        count = count.eval("count = count_x + count_y").reindex(
                            columns, axis=1
                        )
            except:
                self.logger.warning(
                    "Error in count_detail for record {}".format(rec.id)
                )

        return count.groupby(list(columns[:-1])).sum().reset_index()



LOG_ORIGIN = "log_from"
LOG_STEP_LABEL = "step_label"
LOG_RECORD_ID = "record_id"
LOG_STEP_STATUS = "step_status"


class StepStatus:

    INIT = 0
    DATA_READY = 10
    PROCESSING = 20
    DONE = 30
    ERROR = 40
    DATA_NOT_AVIABLE = 41


def record_log_file_path(data_path, rec_id):
    return data_path / "LOG" / "{}.log".format(rec_id)



# coding=utf-8

from .model import Model
from .step import Step
from .step_collection import StepCollection
from .scope import Scope
from .record import Record
from .run_manager import RunManager
from .subset import Subset
from .dataset import Dataset, ConfigurationError


__all__ = [
    "Model",
    "RunManager",
    "Scope",
    "Dataset",
    "ConfigurationError",
    "Subset",
    "Record",
    "Step",
    "StepCollection",
]



# coding=utf-8

import pandas as pd
from sinagot.models import Model, Step
from sinagot.utils import LOG_STEP_LABEL, LOG_STEP_STATUS


class StepCollection(Model):
    """Manage all steps of a category.

    Args:
        model (instance): Parent model of the collection.

    Attributes:
        model (instance): Parent model of the collection.

    """

    _MODEL_TYPE = "step_collection"

    def __init__(self, model):

        super().__init__(model.dataset)
        self.model = model
        self.task = self.model.task
        self.modality = self.model.modality

    @property
    def _scripts_names(self):
        try:
            sn = self.dataset.config["modalities"][self.modality]["scripts"]
        except KeyError:
            return []
        try:
            return (
                sn
                + self.dataset.config["modalities"][self.modality]["tasks_scripts"][
                    self.task
                ]
            )
        except KeyError:
            return sn

    def get_script(self, script_name):
        return self._get_module("Script", self.modality, script_name)

    def get(self, script):

        return Step(script=script, model=self.model)

    def all(self):
        for script_name in self._scripts_names:
            yield self.get(script_name)

    def find(self, pattern):

        for script_name in self._scripts_names:
            if pattern in script_name:
                yield self.get(script_name)

    def first(self):
        """Get the first step.

        Returns:
            instance: Step instance
        """

        return self.get(self._scripts_names[0])

    def count(self):
        return len(self._scripts_names)

    def status(self):
        if self.model._MODEL_TYPE == "record":
            if self.count() > 0:
                return pd.DataFrame(
                    [
                        {
                            "record_id": self.model.id,
                            "task": self.task,
                            "modality": self.modality,
                            "step_index": index,
                            LOG_STEP_LABEL: step.label,
                            LOG_STEP_STATUS: step.status(),
                        }
                        for index, step in zip(range(1, self.count() + 1), self.all())
                    ]
                )
            else:
                return None
        elif self.model.count() == 0:
            return None
        else:
            return pd.concat(
                [rec.status() for rec in self.model.all() if rec.status() is not None]
            )



# coding=utf-8

import re
from pathlib import Path
from typing import Union, Optional
import toml
import pandas as pd
from sinagot.models import Subset, Record, RunManager
from sinagot.config import ConfigurationError
from sinagot.logger import logger_factory

try:
    from sinagot.plugins import DaskRunManager

    dask_enable = True
except ImportError:
    dask_enable = False


class Dataset(Subset):
    """
    Dataset is the main class to handle data.
    It handle all configuration information and objects non specific to a particular data as the logger.
    It's also a [Subset](subset.md) instance used to manipulate a collection off all the records.

    It required a configuration file in toml format.

    Usage:

    ```python
    from sinagot import Dataset
    ds = Dataset(path/to/config.toml)
    ```
    """

    _subscope_class = Subset

    def __init__(
        self,
        config_path: Union[str, Path],
        data_path: Optional[Union[str, Path]] = None,
    ):
        """
        Arguments:
            config_path: path to the config file in toml.
            data_path: path to the folder of the dataset.
        """

        # Load config
        self._config_path = Path(config_path)
        if self._config_path.is_dir():
            config_path = self._config_path / "dataset.toml"
        self.config = toml.load(config_path)
        """Config dictionnary from config file."""

        # Get data_path
        if data_path is None:
            data_path = self.config["path"].get("data", self._config_path)
        self._data_path = self._resolve_path(data_path)

        # Get _scripts_path from config
        try:
            self._scripts_path = self._resolve_path(self.config["path"]["scripts"])
        except KeyError:
            self._scripts_path = self._config_path

        # Init logger
        self.logger = logger_factory(self.config)
        """Logger of the dataset"""

        # Init Run management
        try:
            run_mode = self.config["run"]["mode"]
        except KeyError:
            run_mode = "main_process"
        if run_mode == "dask":
            if not dask_enable:
                raise ConfigurationError("Dask is not enable in this environment")
            run_manager = DaskRunManager
        else:
            run_manager = RunManager

        self._run_manager = run_manager(self)

        #  Init from subset
        super().__init__(self)

        # Set subscope from config to Subset and Record class
        Subset._set_subscopes(self)
        Record._set_subscopes(self)

    def _resolve_path(self, raw_path):
        """Resolve path from config from raw_path and self.config_path if raw path
        is relative to current dir i.e. begin with './' or '...'

        Args:
            raw_path (str or pathlib.Path): path from config to resolve

        Returns:
            pathlib.Path: absolute path
        """

        path = Path(raw_path)
        if re.match(r"(?:\.{1,2}\/.*|\.$)", str(raw_path)):
            path = Path(self._config_path, raw_path).resolve()
        return path



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sinagot', 'sinagot.models', 'sinagot.plugins']

package_data = \
{'': ['*']}

install_requires = \
['json-log-formatter>=0.3.0,<0.4.0',
 'pandas>=1.0.2,<2.0.0',
 'toml>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'sinagot',
    'version': '0.1.1',
    'description': 'Python library to manage data processing on a dataset.',
    'long_description': '# Sinagot\n\nSinagot is a Python library to manage data processing with *scripts* on a *dataset*. \nSinagot is able to batch scripts runs with a simple API. \nParallelization of data processing is possible with [Dask.distributed](https://distributed.dask.org/en/latest/). \n\n## Installation\n\nSinagot is available on PyPi :\n\n```bash\npip install sinagot\n```\n\n## Full Documentation\n\n<https://sinagot.readthedocs.io/en/latest/>\n\n## Concept\n\n*Dataset* are structured around some core concept : **record**, **subset**, **task**, **modality** and **script**. \nA *record*, identified by its unique ID, correspond to a recording session where experimental *tasks* are performed \ngenerating data of various *modalities*. Raw data of a **record** are processed with **scripts** to generate more useful data.\n\n> The idea of Sinagot emerged for the data management of an EEG platform called SoNeTAA. \n> For documentation purpose SoNeTAA dataset structure will be used as example. \n\nOn SoNeTAA, a record with an ID with timestamp info in this format REC-[YYMMDD]-[A-Z], \nfor example "REC-200331-A". \n\nFor a record, 3 tasks are performed : "RS", "MMN" and "HDC",\n2 main modalities handle data for every tasks: "EEG" and "clinical", and a third one "behavior" exists only for HDC.\n\n## Demo with SoNeTAA example \n\n### Create a Dataset instance\n\nImport Dataset class\n\n```python\n>>> from sinagot import Dataset\n```\n\nA `Dataset` instance need 3 things : \n\n- A config file in toml format.\n- A folder containing the dataset\n- A folder containing all the scripts\n\nTo instantiate a dataset use the config file path as argument :\n\n```python\n>>> ds = Dataset(\'/path/to/conf\')\n>>> ds\n<Dataset instance | task: None, modality: None>\n```\n\n> Be sure that dataset path and scripts path are correctly set in the config file\n\n### Explore records\n\nYou can list all records ids :\n\n```python\n>>> for id in ds.ids():\n...     print(id)\n...\nREC-200331-A\nREC-200331-B\n```\n\nCreate a `Record` instance. For a specific record :\n\n```python\n>>> rec = ds.get(\'REC-200331-A\')\n>>> rec\n<Record instance | id: REC-200331-A, task: None, modality: None>\n```\n\nOr the first record found :\n\n```python\n>>> ds.first()\n<Record instance | id: REC-200331-B, task: None, modality: None>\n```\n\n> Records are not sort by their ids.\n\n### run scripts\n\nYou can run all scripts for each record of the dataset :\n\n```python\n>>> ds.run()\n2020-03-31 16:03:58,869 : Begin step run\n...\n2020-03-31 16:03:58,869 : Step run finished\n```\n\nOr for a single record :\n\n```python\n>>> rec.run()\n2020-03-31 16:06:57,313 : Begin step run\n...\n2020-03-31 16:06:57,314 : Step run finished\n```\n\n### Explore by task or modality\n\nEach dataset or record has **subscopes** corresponding to their tasks and modalities\nsimply accessible by self attributes with the scope name.\n\nFor example to select only the task RS of the dataset :\n\n```python\n>>> ds.RS\n<Subset instance | task: RS, modality: None>\n```\n\n> A dataset subscope is a *subset*.\n\nOr the EEG modality of a record :\n\n```python\n>>> rec.EEG\n<Record instance | id: REC-200331-A, task: None, modality: EEG>\n```\n\nYou can select a specific couple of task and modality (called **unit**) :\n\n```python\n>>> ds.RS.EEG\n<Subset instance | task: RS, modality: EEG>\n>>> ds.EEG.RS\n<Subset instance | task: RS, modality: EEG>\n```\n',
    'author': 'Yann Beauxis',
    'author_email': 'dev@yannbeauxis.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/YannBeauxis/sinagot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# coding=utf-8
"""Common functions used for script"""

import logging
from pathlib import Path
from collections import namedtuple
from json_log_formatter import JSONFormatter
from sinagot.utils import (
    record_log_file_path,
    LOG_ORIGIN,
    LOG_STEP_LABEL,
    LOG_RECORD_ID,
    LOG_STEP_STATUS,
    StepStatus,
)


class ScriptTemplate:
    """
        Template class for scripts

        Args:
            data_path (str): Folder path of data corresponding
                to :code:`data_path` attribute of dataset or record models.
            id_ (str): Id of the record or task to run the script.
            task (str): Taks label
    """

    PATH_IN = ("FOLDER_IN", "{id}-{task}.in")
    PATH_OUT = ("FOLDER_OUT", "{id}-{task}.out")

    _logger_file_handler = None
    _logger = None
    logger = None

    def __init__(self, data_path, id_, task, opts={}, logger_namespace=None):

        self.status = StepStatus.INIT
        self.data_path = data_path
        self.id = id_
        self.task = task
        module_split = self.__class__.__module__.split(".")
        self.modality = module_split[-2]
        self.label = module_split[-1]

        self.opts = opts

        self._logger_namespace = logger_namespace

        self._set_path()

    def _init_logger(self):
        file_handler_path = record_log_file_path(self.data_path, self.id)
        file_handler = logging.FileHandler(file_handler_path)
        file_formatter = JSONFormatter()
        file_handler.setFormatter(file_formatter)
        self._logger_file_handler = file_handler
        logger_ = logging.getLogger(self._logger_namespace)
        logger_.setLevel(logging.INFO)
        self._logger = logger_
        logger = logging.LoggerAdapter(logger_, self._log_extra(StepStatus.PROCESSING))
        self.logger = logger

    def _log_extra(self, status):
        return {
            LOG_ORIGIN: "script",
            LOG_RECORD_ID: self.id,
            "scope": "record",
            "task": self.task,
            "modality": self.modality,
            LOG_STEP_LABEL: self.label,
            LOG_STEP_STATUS: status,
        }

    def _log_status(self, message, status):
        self._logger.info(message, extra=self._log_extra(status))

    def _set_path(self):

        ScriptPath = namedtuple("IOPath", ["input", "output"])
        self.path = ScriptPath(
            input=self._get_path(self.PATH_IN), output=self._get_path(self.PATH_OUT),
        )

    def _get_path(self, raw_path):

        str_path = (
            rp.format(id=self.id, task=self.task, **self.opts) for rp in raw_path
        )
        return Path(self.data_path, *str_path)

    @property
    def data_exist(self):
        DataStatus = namedtuple("IOPath", ["input", "output"])

        def path_exist(path):
            if isinstance(path, Path):
                return path.exists()
            elif isinstance(path, dict):
                return all([path_exist(p) for p in path.values()])

        return DataStatus(*(path_exist(path) for path in self.path))

    def _run(self, force=False):
        """Required run function. Called by step model :code:`run()` method."""

        self._init_logger()
        self._logger.addHandler(self._logger_file_handler)
        self._log_status("Init run", StepStatus.INIT)
        if not self.data_exist.input:
            self._log_status("Data not aviable", StepStatus.DATA_NOT_AVIABLE)
        elif (not force) and self.data_exist.output:
            self._log_status("Run already processed", StepStatus.DONE)
        else:
            self.status = StepStatus.PROCESSING
            self._log_status("Processing run", StepStatus.PROCESSING)
            self.run()
            self._log_status("Run finished", StepStatus.DONE)

        self._logger.removeHandler(self._logger_file_handler)

        return True

    def _run_force(self):
        self._run(force=True)

    def run(self):
        self.logger.info("Running ...")
